
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.asm;

import static org.objectweb.asm.Opcodes.*;

import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;

import de.unkrig.commons.nullanalysis.Nullable;

/**
 * Utility class around {@link org.objectweb.asm.Opcodes}.
 */
public final
class OpcodeUtil {

    private
    OpcodeUtil() {}

    private static final String[] OPCODE_NAMES = {
        "NOP", "ACONST_NULL", "ICONST_M1", "ICONST_0", "ICONST_1", "ICONST_2", "ICONST_3", "ICONST_4",
        "ICONST_5", "LCONST_0", "LCONST_1", "FCONST_0", "FCONST_1", "FCONST_2", "DCONST_0", "DCONST_1 = 15",
        "BIPUSH", "SIPUSH", "LDC", "LDC_W", "LDC2_W", "ILOAD", "LLOAD", "FLOAD",
        "DLOAD", "ALOAD", "ILOAD_0", "ILOAD_1", "ILOAD_2", "ILOAD_3", "LLOAD_0", "LLOAD_1",
        "LLOAD_2", "LLOAD_3", "FLOAD_0", "FLOAD_1", "FLOAD_2", "FLOAD_3", "DLOAD_0", "DLOAD_1",
        "DLOAD_2", "DLOAD_3", "ALOAD_0", "ALOAD_1", "ALOAD_2", "ALOAD_3", "IALOAD", "LALOAD",
        "FALOAD", "DALOAD", "AALOAD", "BALOAD", "CALOAD", "SALOAD", "ISTORE", "LSTORE",
        "FSTORE", "DSTORE", "ASTORE", "ISTORE_0", "ISTORE_1", "ISTORE_2", "ISTORE_3", "LSTORE_0",
        "LSTORE_1", "LSTORE_2", "LSTORE_3", "FSTORE_0", "FSTORE_1", "FSTORE_2", "FSTORE_3", "DSTORE_0",
        "DSTORE_1", "DSTORE_2", "DSTORE_3", "ASTORE_0", "ASTORE_1", "ASTORE_2", "ASTORE_3", "IASTORE",
        "LASTORE", "FASTORE", "DASTORE", "AASTORE", "BASTORE", "CASTORE", "SASTORE", "POP",
        "POP2", "DUP", "DUP_X1", "DUP_X2", "DUP2", "DUP2_X1", "DUP2_X2", "SWAP",
        "IADD", "LADD", "FADD", "DADD", "ISUB", "LSUB", "FSUB", "DSUB",
        "IMUL", "LMUL", "FMUL", "DMUL", "IDIV", "LDIV", "FDIV", "DDIV",
        "IREM", "LREM", "FREM", "DREM", "INEG", "LNEG", "FNEG", "DNEG",
        "ISHL", "LSHL", "ISHR", "LSHR", "IUSHR", "LUSHR", "IAND", "LAND",
        "IOR", "LOR", "IXOR", "LXOR", "IINC", "I2L", "I2F", "I2D",
        "L2I", "L2F", "L2D", "F2I", "F2L", "F2D", "D2I", "D2L",
        "D2F", "I2B", "I2C", "I2S", "LCMP", "FCMPL", "FCMPG", "DCMPL",
        "DCMPG", "IFEQ", "IFNE", "IFLT", "IFGE", "IFGT", "IFLE", "IF_ICMPEQ",
        "IF_ICMPNE", "IF_ICMPLT", "IF_ICMPGE", "IF_ICMPGT", "IF_ICMPLE", "IF_ACMPEQ", "IF_ACMPNE", "GOTO",
        "JSR", "RET", "TABLESWITCH", "LOOKUPSWITCH", "IRETURN", "LRETURN", "FRETURN", "DRETURN",
        "ARETURN", "RETURN", "GETSTATIC", "PUTSTATIC", "GETFIELD", "PUTFIELD", "INVOKEVIRTUAL", "INVOKESPECIAL",
        "INVOKESTATIC", "INVOKEINTERFACE", "INVOKEDYNAMIC", "NEW", "NEWARRAY", "ANEWARRAY", "ARRAYLENGTH", "ATHROW",
        "CHECKCAST", "INSTANCEOF", "MONITORENTER", "MONITOREXIT", "WIDE", "MULTIANEWARRAY", "IFNULL", "IFNONNULL",
        "GOTO_W", "JSR_W", null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null,
    };

    public static String
    getName(AbstractInsnNode ain) {
        return OPCODE_NAMES[ain.getOpcode()];
    }

    /**
     * Description of a Java&trade; bytecode instruction.
     */
    public static final
    class InsnDescription {

        /**
         * The types of the instructions's operands.
         */
        @Nullable public final Type[] argumentOperandTypes;

        /**
         * The type of the instruction's result operand (or {@code null}).
         */
        @Nullable public final Type resultOperandType;

        public
        InsnDescription(@Nullable Type[] argumentOperandTypes, @Nullable Type resultOperandType) {
            this.argumentOperandTypes = argumentOperandTypes;
            this.resultOperandType    = resultOperandType;
        }
    }

    private static final InsnDescription[] INSN_DESCRIPTIONS = new InsnDescription[256];
    static { initDescriptions(); }
    private static void
    initDescriptions() {
        addRInsn(Type.INT_TYPE,     ILOAD);
        addRInsn(Type.LONG_TYPE,    LLOAD);
        addRInsn(Type.FLOAT_TYPE,   FLOAD);
        addRInsn(Type.DOUBLE_TYPE,  DLOAD);
        addRInsn(Types.OBJECT_TYPE, ALOAD);

        addAInsn(Type.INT_TYPE,     ISTORE);
        addAInsn(Type.LONG_TYPE,    LSTORE);
        addAInsn(Type.FLOAT_TYPE,   FSTORE);
        addAInsn(Type.DOUBLE_TYPE,  DSTORE);
        addAInsn(Types.OBJECT_TYPE, ASTORE);

        addAARInsn(Types.INT_ARRAY_TYPE,       Type.INT_TYPE, Type.INT_TYPE,     IALOAD);
        addAARInsn(Types.LONG_ARRAY_TYPE,      Type.INT_TYPE, Type.LONG_TYPE,    LALOAD);
        addAARInsn(Types.FLOAT_ARRAY_TYPE,     Type.INT_TYPE, Type.FLOAT_TYPE,   FALOAD);
        addAARInsn(Types.DOUBLE_ARRAY_TYPE,    Type.INT_TYPE, Type.DOUBLE_TYPE,  DALOAD);
        addAARInsn(Types.REFERENCE_ARRAY_TYPE, Type.INT_TYPE, Types.OBJECT_TYPE, AALOAD);
        addAARInsn(Types.BYTE_ARRAY_TYPE,      Type.INT_TYPE, Type.BYTE_TYPE,    BALOAD);
        addAARInsn(Types.CHAR_ARRAY_TYPE,      Type.INT_TYPE, Type.CHAR_TYPE,    CALOAD);
        addAARInsn(Types.SHORT_ARRAY_TYPE,     Type.INT_TYPE, Type.SHORT_TYPE,   SALOAD);

        addAAAInsn(Types.INT_ARRAY_TYPE,       Type.INT_TYPE, Type.INT_TYPE,     IASTORE);
        addAAAInsn(Types.LONG_ARRAY_TYPE,      Type.INT_TYPE, Type.LONG_TYPE,    LASTORE);
        addAAAInsn(Types.FLOAT_ARRAY_TYPE,     Type.INT_TYPE, Type.LONG_TYPE,    FASTORE);
        addAAAInsn(Types.DOUBLE_ARRAY_TYPE,    Type.INT_TYPE, Type.DOUBLE_TYPE,  DASTORE);
        addAAAInsn(Types.REFERENCE_ARRAY_TYPE, Type.INT_TYPE, Types.OBJECT_TYPE, AASTORE);
        addAAAInsn(Types.BYTE_ARRAY_TYPE,      Type.INT_TYPE, Type.BYTE_TYPE,    BASTORE);
        addAAAInsn(Types.CHAR_ARRAY_TYPE,      Type.INT_TYPE, Type.CHAR_TYPE,    CASTORE);
        addAAAInsn(Types.SHORT_ARRAY_TYPE,     Type.INT_TYPE, Type.SHORT_TYPE,   SASTORE);

        addAARInsn(Type.INT_TYPE,    Type.INT_TYPE,    Type.INT_TYPE,    IADD, ISUB, IMUL, IDIV, IREM, IAND, IOR, IXOR);
        addAARInsn(Type.LONG_TYPE,   Type.LONG_TYPE,   Type.LONG_TYPE,   LADD, LSUB, LMUL, LDIV, LREM, LAND, LOR, LXOR);
        addAARInsn(Type.FLOAT_TYPE,  Type.FLOAT_TYPE,  Type.FLOAT_TYPE,  FADD, FSUB, FMUL, FDIV, FREM);
        addAARInsn(Type.DOUBLE_TYPE, Type.DOUBLE_TYPE, Type.DOUBLE_TYPE, DADD, DSUB, DMUL, DDIV, DREM);

        addAARInsn(Type.LONG_TYPE,   Type.LONG_TYPE,   Type.INT_TYPE, LCMP);
        addAARInsn(Type.FLOAT_TYPE,  Type.FLOAT_TYPE,  Type.INT_TYPE, FCMPG, FCMPL);
        addAARInsn(Type.DOUBLE_TYPE, Type.DOUBLE_TYPE, Type.INT_TYPE, DCMPG, DCMPL);

        addARInsn(Type.INT_TYPE,    Type.INT_TYPE,     INEG);
        addARInsn(Type.LONG_TYPE,   Type.LONG_TYPE,    LNEG);
        addARInsn(Type.FLOAT_TYPE,  Type.FLOAT_TYPE,   FNEG);
        addARInsn(Type.DOUBLE_TYPE, Type.DOUBLE_TYPE,  DNEG);

        addAARInsn(Type.INT_TYPE,  Type.INT_TYPE, Type.INT_TYPE,  ISHL, ISHR, IUSHR);
        addAARInsn(Type.LONG_TYPE, Type.INT_TYPE, Type.LONG_TYPE, LSHL, LSHR, LUSHR);

        addAAInsn(Type.INT_TYPE,     Type.INT_TYPE,     IF_ICMPEQ, IF_ICMPNE, IF_ICMPLT, IF_ICMPGE, IF_ICMPGT, IF_ICMPLE); // SUPPRESS CHECKSTYLE LineLength
        addAAInsn(Types.OBJECT_TYPE, Types.OBJECT_TYPE, IF_ACMPEQ, IF_ACMPNE);

        addARInsn(Type.INT_TYPE,    Type.LONG_TYPE,   I2L);
        addARInsn(Type.INT_TYPE,    Type.FLOAT_TYPE,  I2F);
        addARInsn(Type.INT_TYPE,    Type.DOUBLE_TYPE, I2D);
        addARInsn(Type.INT_TYPE,    Type.BYTE_TYPE,   I2B);
        addARInsn(Type.INT_TYPE,    Type.CHAR_TYPE,   I2C);
        addARInsn(Type.INT_TYPE,    Type.SHORT_TYPE,  I2S);
        addARInsn(Type.LONG_TYPE,   Type.INT_TYPE,    L2I);
        addARInsn(Type.LONG_TYPE,   Type.FLOAT_TYPE,  L2F);
        addARInsn(Type.LONG_TYPE,   Type.DOUBLE_TYPE, L2D);
        addARInsn(Type.FLOAT_TYPE,  Type.INT_TYPE,    F2I);
        addARInsn(Type.FLOAT_TYPE,  Type.LONG_TYPE,   F2L);
        addARInsn(Type.FLOAT_TYPE,  Type.DOUBLE_TYPE, F2D);
        addARInsn(Type.DOUBLE_TYPE, Type.INT_TYPE,    D2I);
        addARInsn(Type.DOUBLE_TYPE, Type.LONG_TYPE,   D2L);
        addARInsn(Type.DOUBLE_TYPE, Type.FLOAT_TYPE,  D2F);

        addAInsn(Type.INT_TYPE, TABLESWITCH, LOOKUPSWITCH);

        addAInsn(Type.INT_TYPE,     IRETURN);
        addAInsn(Type.LONG_TYPE,    LRETURN);
        addAInsn(Type.FLOAT_TYPE,   FRETURN);
        addAInsn(Type.DOUBLE_TYPE,  DRETURN);
        addAInsn(Types.OBJECT_TYPE, ARETURN);
        addInsn(RETURN);
    }
    private static void
    addInsn(int... opcodes) {
        addInsns(new InsnDescription(null, null), opcodes);
    }
    private static void
    addAInsn(Type argumentOperandType, int... opcodes) {
        addInsns(new InsnDescription(new Type[] { argumentOperandType }, null), opcodes);
    }
    private static void
    addARInsn(Type argumentOperandType, Type resultOperandType, int... opcodes) { // SUPPRESS CHECKSTYLE AbbreviationCheck
        addInsns(new InsnDescription(new Type[] { argumentOperandType }, resultOperandType), opcodes);
    }
    private static void
    addAAInsn(Type argumentOperandType1, Type argumentOperandType2, int... opcodes) { // SUPPRESS CHECKSTYLE AbbreviationCheck
        addInsns(new InsnDescription(new Type[] { argumentOperandType1, argumentOperandType2 }, null), opcodes);
    }
    private static void
    addAARInsn(Type argumentOperandType1, Type argumentOperandType2, Type resultOperandType, int... opcodes) { // SUPPRESS CHECKSTYLE AbbreviationCheck
        addInsns(new InsnDescription(new Type[] { argumentOperandType1, argumentOperandType2 }, resultOperandType), opcodes);
    }
    private static void
    addAAAInsn(Type argumentOperandType1, Type argumentOperandType2, Type argumentOperandType3, int... opcodes) { // SUPPRESS CHECKSTYLE AbbreviationCheck
        addInsns(new InsnDescription(new Type[] { argumentOperandType1, argumentOperandType2, argumentOperandType3 }, null), opcodes);
    }
    private static void
    addRInsn(Type resultOperandType, int... opcodes) {
        addInsns(new InsnDescription(null, resultOperandType), opcodes);
    }
    private static void
    addInsns(InsnDescription insnDescription, int... opcodes) {
        for (int opcode : opcodes) {
            assert INSN_DESCRIPTIONS[opcode] == null;
            INSN_DESCRIPTIONS[opcode] = insnDescription;
        }
    }

    /**
     * @return Metainformation about the given {@code opcode}
     */
    public static InsnDescription
    getInsnDescription(int opcode) {
        return INSN_DESCRIPTIONS[opcode];
    }
}
