
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.fileprocessing;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.zip.ZipEntry;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.contentsprocessing.ContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.GzipContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.ZipContentsProcessor;
import de.unkrig.commons.file.zipentryprocessing.SelectiveZipEntryProcessor;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryContentsProcessor;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryProcessor;
import de.unkrig.commons.lang.protocol.Predicate;

/**
 * Various file processing utility methods.
 */
public final
class FileProcessingUtil {

    private
    FileProcessingUtil() {}

    /**
     * @see #process(List, FileProcessor, ExceptionHandler)
     */
    public static void
    process(List<File> files, FileProcessor fileProcessor) throws IOException {
        process(files, fileProcessor, ExceptionHandler.<IOException>defaultHandler());
    }

    /**
     * Invokes the {@code fileProcessor} for each of the {@code files}.
     *
     * @param exceptionHandler Called if the processing of one of the {@code inputFiles} throws an {@link IOException}
     *                         or a {@link RuntimeException}
     */
    public static void
    process(
        List<File>                    files,
        FileProcessor                 fileProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) throws IOException {
        for (File file : files) {
            try {
                fileProcessor.process(file);
            } catch (IOException ioe) {
                exceptionHandler.handle(wrap("Processing '" + file + "'", ioe));
            } catch (RuntimeException re) {
                exceptionHandler.handle(wrap("Processing '" + file + "'", re));
            }
        }
    }

    /**
     * @see #newZipFileProcessor(boolean, Predicate, ContentsProcessor, ExceptionHandler)
     */
    public static FileProcessor
    newZipFileProcessor(
        boolean           lookIntoZip,
        boolean           lookIntoNestedZip,
        boolean           lookIntoGzip,
        boolean           lookIntoNestedGzip,
        Predicate<String> pathPredicate,
        ContentsProcessor contentsProcessor
    ) {
        return newZipFileProcessor(
            lookIntoZip,
            lookIntoNestedZip,
            lookIntoGzip,
            lookIntoNestedGzip,
            pathPredicate,
            contentsProcessor,
            ExceptionHandler.<IOException>defaultHandler()
        );
    }

    /**
     * Returns a {@link FileProcessor} that calls the {@code contentsProcessor}, except if {@code lookIntoZip} is
     * {@code true} and the file is in ZIP format. In the latter case, the {@code contentsProcessor} in invoked for
     * each entry in the ZIP file that matches the {@code pathPredicate}, except for those which are in ZIP format;
     * these are processed recursively (if {@code lookIntoNestedZip} is true).
     * <p>
     * This method optimizes the process to the effect that the ZIP file (and nested ZIP files) are examined iff
     * it could contain elements that match the {@code pathPredicate}.
     * <pre>
     *                |
     *                v
     *            SelFP:name-----+
     *             y|       n    |
     *              |            v
     *              |          SelFP:name!
     *              |           y|     |n
     *              |            v     x
     *             (2)          (2)
     *           CP| |dfltCP  CP| |dfltCP
     *             v v          | x
     *          delegateCP <----+
     * </pre>
     * (2) is the object returned by {@link #newZipFileProcessor(boolean, Predicate, ContentsProcessor,
     * ContentsProcessor, ExceptionHandler)}.
     */
    public static FileProcessor
    newZipFileProcessor(
        boolean                       lookIntoZip,
        boolean                       lookIntoNestedZip,
        boolean                       lookIntoGzip,
        boolean                       lookIntoNestedGzip,
        final Predicate<String>       pathPredicate,
        ContentsProcessor             contentsProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        if (!lookIntoZip) {
            return new FileContentsProcessor(
                lookIntoGzip ? new GzipContentsProcessor(contentsProcessor, contentsProcessor) : contentsProcessor
            );
        }

        FileProcessor sub1 = newZipFileProcessor(
            lookIntoNestedZip,
            lookIntoGzip,
            lookIntoNestedGzip,
            pathPredicate,
            contentsProcessor,
            contentsProcessor,
            exceptionHandler
        );
        
        FileProcessor sub2 = newZipFileProcessor(
            lookIntoNestedZip,
            lookIntoGzip,
            lookIntoNestedGzip,
            pathPredicate,
            contentsProcessor,
            ContentsProcessor.NOP,
            exceptionHandler
        );

        FileProcessor sel2 = new SelectiveFileProcessor(
            stringPredicateSuffixed(pathPredicate, "!"),
            sub2,
            FileProcessor.NOP
        );
        
        return new SelectiveFileProcessor(
            pathPredicate,
            sub1,
            sel2
        );
    }

    private static Predicate<String>
    stringPredicateSuffixed(final Predicate<String> delegate, final String suffix) {
        return new Predicate<String>() {

            @Override public boolean
            evaluate(String subject) { return delegate.evaluate(subject + suffix); }
        };
    }
    
    /**
     * Returns a {@link FileProcessor} that calls the {@code defaultContentsProcessor} for every file, except if the
     * file is in ZIP format. In the latter case, the {@code contentsProcessor} is invoked for each entry in the ZIP
     * file that matches the {@code pathPredicate}, except for those which are in ZIP format; these are processed
     * recursively iff {@code lookIntoNestedZip} is true.
     * <pre>
     *                   |
     *                   v
     *                ZipFP------------+
     *               ZE|     dflt      |
     *                 v               v
     *      +---->SelZEH:name        GzipCP
     *      |     y|       |n         | |dflt
     *      |      |       v          | |
     *      |      |     SelZEH:name! | |
     *      |      |        y|    |n  | |
     *      |   GzipCP       |    x   | |
     *      |      |      GzipCP      | |
     *      |      v         |        | |
     *      |      +<-----------------+ |
     *      |      |         |          |
     *      |      v         v          |
     *      |    ZipCP     ZipCP        |
     *      |   ZE| |dflt ZE| |dflt     |
     *      +-----+-|-------+ x         |
     *              v                   v
     *    contentsProcessor  defaultContentsProcessor
     * </pre>
     */
    private static FileProcessor
    newZipFileProcessor(
        boolean                       lookIntoNestedZip,
        boolean                       lookIntoGzip,
        boolean                       lookIntoNestedGzip,
        Predicate<String>             pathPredicate,
        ContentsProcessor             contentsProcessor,
        ContentsProcessor             defaultContentsProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {

        ZipEntryProcessor sel1;
        ContentsProcessor zipcp1;

        if (!lookIntoNestedZip) {
            sel1 = new SelectiveZipEntryProcessor(
                pathPredicate,
                new ZipEntryContentsProcessor(
                    lookIntoNestedGzip
                    ? new GzipContentsProcessor(contentsProcessor, contentsProcessor)
                    : contentsProcessor
                ),
                ZipEntryProcessor.NOP
            );
            zipcp1 = contentsProcessor;
        } else {
            final ZipEntryProcessor[] loopback   = new ZipEntryProcessor[1];
            ZipEntryProcessor         toLoopback = new ZipEntryProcessor() {
                
                @Override public void
                process(ZipEntry entry, String path, InputStream inputStream) throws IOException {
                    loopback[0].process(entry, path, inputStream);
                }
            };
    
            zipcp1 = new ZipContentsProcessor(
                pathPredicate,     // nestedZipPathPredicate
                toLoopback,        // nestedZipEntryProcessor
                contentsProcessor, // defaultContentsProcessor
                exceptionHandler   // exceptionHandler
            );
            
            ContentsProcessor zipcp2 = new ZipContentsProcessor(
                pathPredicate,         // nestedZipPathPredicate
                toLoopback,            // nestedZipEntryProcessor
                ContentsProcessor.NOP, // defaultContentsProcessor
                exceptionHandler       // exceptionHandler
            );
            if (lookIntoNestedGzip) zipcp2 = new GzipContentsProcessor(zipcp2, zipcp2);
    
            ZipEntryProcessor sel2 = new SelectiveZipEntryProcessor(
                stringPredicateSuffixed(pathPredicate, "!"),
                new ZipEntryContentsProcessor(lookIntoNestedGzip ? new GzipContentsProcessor(zipcp2, zipcp2) : zipcp2),
                ZipEntryProcessor.NOP
            );
            
            sel1 = (loopback[0] = new SelectiveZipEntryProcessor(
                pathPredicate,
                new ZipEntryContentsProcessor(lookIntoNestedGzip ? new GzipContentsProcessor(zipcp1, zipcp1) : zipcp1),
                sel2
            ));
        }

        ContentsProcessor gzcp = (
            lookIntoGzip
            ? new GzipContentsProcessor(zipcp1, defaultContentsProcessor)
            : defaultContentsProcessor
        );

        return new ZipFileProcessor(
            sel1,                            // zipEntryProcessor
            new FileContentsProcessor(gzcp), // defaultFileProcessor
            exceptionHandler
        );
    }

    /**
     * @see #newDirectoryTreeProcessor(Predicate, FileProcessor, ExceptionHandler)
     */
    public static FileProcessor
    newDirectoryTreeProcessor(
        Predicate<String> defaultFileProcessor,
        FileProcessor     fileProcessor
    ) {
        return newDirectoryTreeProcessor(
            defaultFileProcessor,
            fileProcessor,
            ExceptionHandler.<IOException>defaultHandler()
        );
    }

    /**
     * Creates a file processor that recurses through the directory tree, applies the {@code directoryPathPredicate},
     * and eventually calls the delegate file processor.
     * <pre>
     *             |
     *             v
     *   +---->SelFP:name
     *   |     y|      |n
     *   |      |      v
     *   |      |   SelFP:name/
     *   |      |    y|    |n
     *   |      |     |    x
     *   |      v     v
     *   |   DirFP  DirFP
     *   |   y| |n  y| |n
     *   +----+-|----+ x
     *          v
     *       delegate
     * </pre>
     */
    public static FileProcessor
    newDirectoryTreeProcessor(
        Predicate<String>             pathPredicate,
        FileProcessor                 delegate,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        final FileProcessor[] loopback   = new FileProcessor[1];
        FileProcessor         toLoopback = new FileProcessor() {
            @Override public void process(File file) throws IOException { loopback[0].process(file); }
        };

        FileProcessor dir1 = new DirectoryProcessor(
            toLoopback,      // memberProcessor
            delegate,        // defaultFileProcessor
            exceptionHandler // exceptionHandler
        );
        
        FileProcessor dir2 = new DirectoryProcessor(
            toLoopback,        // memberProcessor
            FileProcessor.NOP, // defaultFileProcessor
            exceptionHandler   // exceptionHandler
        );

        FileProcessor sel2 = new SelectiveFileProcessor(
            stringPredicateSuffixed(pathPredicate, "/"), // pathPredicate
            dir2,                                        // delegate1
            FileProcessor.NOP                            // delegate2
        );
        
        FileProcessor sel1 = (loopback[0] = new SelectiveFileProcessor(
            pathPredicate, // pathPredicate
            dir1,          // delegate1
            sel2           // delegate2
        ));

        return sel1;
    }
}
