
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import java.io.File;
import java.io.IOException;

import de.unkrig.commons.file.contentstransformation.ContentsTransformerUtil;

/**
 * @see #transform(File, File)
 */
public
interface FileTransformer {

    /**
     * Reads plain file {@code in}, transforms its contents, creates file {@code out}, and stores the transformed
     * contents there.
     */
    void
    transform(File in, File out) throws IOException;

    /**
     * If {@code in} equals {@code out} ("in-place transformation"), then nothing is done. Otherwise ("out-of-place
     * transformation") an exact copy of {@code in} named {@code out} is created.
     */
    FileTransformer UNCHANGED = new FileTransformer() {

        @Override public void
        transform(File in, File out) throws IOException {
            if (in.equals(out)) {
                ;
            } else {
                new FileContentsTransformer(ContentsTransformerUtil.copy()).transform(in, out);
            }
        }
    };

    /**
     * Does nothing. If {@code in} equals {@code out} ("in-place transformation"), then the {@code in} file is not
     * touched. Otherwise ("out-of-place transformation") nothing is done, i.e. a file {@code out} is <i>not</i>
     * created.
     */
    FileTransformer NOP = new FileTransformer() {

        @Override public void
        transform(File in, File out) {
            ;
        }
    };

    /**
     * If {@code in} equals {@code out} ("in-place transformation"), then file {@code in} is deleted. Otherwise
     * ("out-of-place transformation") nothing is done, i.e. a file {@code out} is <i>not</i> created.
     */
    FileTransformer REMOVE = new FileTransformer() {

        @Override public void
        transform(File in, File out) throws IOException {
            if (in.equals(out)) {
                if (!in.delete()) throw new IOException("Could not delete '" + in + "'");
            }
        }
    };
}
