
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.io;

import static java.util.logging.Level.SEVERE;
import static java.util.logging.Level.WARNING;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.PipedReader;
import java.io.PipedWriter;
import java.io.Reader;
import java.io.Writer;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.logging.Logger;

import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.lang.protocol.RunnableWhichThrows;
import de.unkrig.commons.lang.protocol.Transformer;
import de.unkrig.commons.nullanalysis.NotNullByDefault;
import de.unkrig.commons.nullanalysis.Nullable;

/**
 * Various {@code java.io}-related utility methods.
 */
public final
class IoUtil {

    private static final Logger LOGGER = Logger.getLogger(IoUtil.class.getName());

    private
    IoUtil() {}

    /**
     * Reads the input stream until end-of-input and writes all data to the output stream. Closes neither of the two
     * streams.
     *
     * @return The number of bytes copied
     */
    public static long
    copy(InputStream in, OutputStream out) throws IOException {
        byte[] buffer = new byte[4096];
        long   count  = 0L;
        try {
            for (;;) {
                int n = in.read(buffer);
                if (n == -1) break;
                out.write(buffer, 0, n);
                count += n;
            }
            out.flush();
        } catch (IOException ioe) {
            throw ExceptionUtil.wrap(count + " bytes copied so far", ioe);
        }
        return count;
    }

    /**
     * Creates and returns a {@link RunnableWhichThrows} that copies bytes from {@code in} to {@code out} until
     * end-of-input.
     */
    public static RunnableWhichThrows<IOException>
    copyRunnable(final InputStream in, final OutputStream out) {
        return new RunnableWhichThrows<IOException>() {

            @Override public void
            run() throws IOException {
                copy(in, out);
            }
        };
    }

    /**
     * Reads the reader until end-of-input and writes all data to the writer. Closes neither the reader nor the writer.
     */
    public static void
    copy(Reader r, Writer w) throws IOException {
        char[] buffer = new char[4096];
        for (;;) {
            int n = r.read(buffer);
            if (n == -1) break;
            w.write(buffer, 0, n);
        }
    }

    /**
     * Creates and returns an {@link OutputStream} that delegates all work to the given {@code delegates}:
     * <ul>
     *   <li>
     *     The {@link OutputStream#write(byte[], int, int) write()} methods write the given data to all the delegates;
     *     if any of these throw an {@link IOException}, it is rethrown, and it is undefined whether all the data was
     *     written to all the delegates.
     *   </li>
     *   <li>
     *     {@link OutputStream#flush() flush()} flushes the delegates; throws the first {@link IOException} that any
     *     of the delegates throws.
     *   </li>
     *   <li>
     *     {@link OutputStream#close() close()} attempts to close <i>all</i> the {@code delegates}; if any of these
     *     throw {@link IOException}s, one of them is rethrown.
     *   </li>
     * </ul>
     */
    @NotNullByDefault(false) public static OutputStream
    tee(final OutputStream... delegates) {
        return new OutputStream() {

            @Override public void
            close() throws IOException {
                IOException caughtIOException = null;
                for (OutputStream delegate : delegates) {
                    try {
                        delegate.close();
                    } catch (IOException ioe) {
                        caughtIOException = ioe;
                    }
                }
                if (caughtIOException != null) throw caughtIOException;
            }

            @Override public void
            flush() throws IOException {
                for (OutputStream delegate : delegates) delegate.flush();
            }

            @Override public void
            write(byte[] b, int off, int len) throws IOException {
                // Overriding this method is not strictly necessary, because "OutputStream.write(byte[], int, int)"
                // calls "OutputStream.write(int)", but "delegate.write(byte[], int, int)" is probably more
                // efficient. However, the behavior is different when one of the delegates throws an exception
                // while being written to.
                for (OutputStream delegate : delegates) delegate.write(b, off, len);
            }

            @Override public void
            write(int b) throws IOException {
                for (OutputStream delegate : delegates) delegate.write(b);
            }
        };
    }

    /**
     * An entity which writes characters to a {@link Writer}.
     */
    public
    interface WritingRunnable {

        /**
         * @see WritingRunnable
         */
        void
        run(Writer w) throws Exception;
    }

    private static final ExecutorService EXECUTOR_SERVICE = new ScheduledThreadPoolExecutor(
        3 * Runtime.getRuntime().availableProcessors()
    );

    /**
     * Executes the {@code writingRunnables} in parallel, <i>concatenates</i> their output, and writes it to the {@code
     * writer}, i.e. the output of the runnables does not mix, but the <i>complete</i> output of the first runnable
     * appears before that of the second runnable, and so on.
     * <p>
     * Since the character buffer for each {@link WritingRunnable} has a limited size, the runnables with higher
     * indexes tend to block if the runnables with lower indexes do not complete quickly enough.
     */
    public static void
    parallel(WritingRunnable[] writingRunnables, final Writer writer) {
        List<Callable<Void>> callables = toCallables(writingRunnables, writer);

        try {
            EXECUTOR_SERVICE.invokeAll(callables);
        } catch (InterruptedException ie) {
            Thread.currentThread().interrupt(); // Preserve interrupt status.
        }
    }

    /**
     * Creates and returns a list of callables; when all of these have been called, then all the given {@code
     * writingRunnables} have been run, and their output is written strictly sequentially to the given {@code writer},
     * even if the callables were called out-of-sequence or in parallel.
     * <p>
     * Deadlocks may occur if lower-index {@code writingRunnables} "depend" on higher-index {@code writingRunnables},
     * i.e. the former do not complete because they wait for a certain state of completion of the latter.
     */
    private static List<Callable<Void>>
    toCallables(WritingRunnable[] writingRunnables, final Writer writer) {
        List<Callable<Void>> callables = new ArrayList<Callable<Void>>(writingRunnables.length + 1);
        final List<Reader>   readers   = new ArrayList<Reader>(writingRunnables.length);

        // Create the 'collector' that concatenates the runnnables' outputs.
        callables.add(new Callable<Void>() {

            @Override @Nullable public Void
            call() throws Exception {
                for (Reader reader : readers) {
                    IoUtil.copy(reader, writer);
                }
                return null;
            }
        });

        for (final WritingRunnable wr : writingRunnables) {

            // Create a PipedReader/PipedWriter pair for communication between the runnable and the 'collector'.
            final PipedWriter pw = new PipedWriter();
            try {
                readers.add(new PipedReader(pw));
            } catch (IOException ioe) {
                throw ExceptionUtil.wrap(
                    "Should never throw an IOException if the argument is a 'fresh' PipedWriter",
                    ioe,
                    AssertionError.class
                );
            }

            // Create a callable that will run the runnable.
            callables.add(new Callable<Void>() {

                @Override @Nullable public Void
                call() throws Exception {
                    try {
                        wr.run(pw);
                        return null;
                    } catch (Exception e) {
                        LOGGER.log(WARNING, null, e);
                        throw e;
                    } catch (Error e) { // SUPPRESS CHECKSTYLE Illegal Catch
                        LOGGER.log(SEVERE, null, e);
                        throw e;
                    } finally {
                        try { pw.close(); } catch (Exception e) {}
                    }
                }
            });
        }
        return callables;
    }

    /**
     * @return All bytes that the given {@link InputStream} produces
     */
    public static byte[]
    readAll(InputStream is) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        copy(is, baos);
        return baos.toByteArray();
    }

    /**
     * An {@link InputStream} that produces zero bytes.
     */
    public static final InputStream EMPTY_INPUT_STREAM = new InputStream() {

        @Override public int
        read() { return -1; }
    };

    /**
     * @return All characters that the given {@link Reader} produces
     */
    public static String
    readAll(Reader r) throws IOException {
        char[]        chars = new char[4096];
        StringBuilder sb    = new StringBuilder();
        for (;;) {
            int n = r.read(chars);
            if (n == -1) break;
            sb.append(chars, 0, n);
        }
        return sb.toString();
    }

    /**
     * @param blockingRead  If {@code false}, then reading the input streams causes a {@link BlockingException} iff no
     *                      data is available.
     * @param blockingWrite If {@code false}, then writing the output streams causes a {@link BlockingException} iff
     *                      there is not enough space in the buffer. Notice that this is ALWAYS the case if the number
     *                      of bytes to write is larger than the buffer's capacity.
     */
    public static Pipe
    byteBufferPipe(final ByteBuffer byteBuffer, final boolean blockingRead, final boolean blockingWrite) {

        /**
         * A {@link ByteBuffer}-based implementation of a {@link Pipe}.
         */
        class ByteBufferPipe implements Pipe {

            private final InputStream  inputStream;
            private final OutputStream outputStream;
            private int                head; // Modified only by the InputStream.
            private int                tail; // Modified only by the OutputStream.
            private boolean            empty = true;
            private boolean            closed;
            private final int          capacity = byteBuffer.capacity();
            private final Object       notFull  = new Object();
            private final Object       notEmpty = new Object();

            @Override public InputStream
            getInputStream()  { return this.inputStream; }

            @Override public OutputStream
            getOutputStream() { return this.outputStream; }

            {
                this.inputStream = new InputStream() {

                    @Override public int
                    read() throws IOException {
                        byte[] ba = new byte[1];
                        return this.read(ba, 0, 1) == -1 ? -1 : 0xff & ba[0];
                    }

                    @NotNullByDefault(false) @Override public int
                    read(byte[] ba, int off, int len) throws IOException {
                        return ByteBufferPipe.this.read(ba, off, len);
                    }

                    @Override public int
                    available() {
                        return ByteBufferPipe.this.getAllocated();
                    }

                    @Override public void
                    close() {
                        ByteBufferPipe.this.empty  = true;
                        ByteBufferPipe.this.closed = true;
                        ByteBufferPipe.this.head   = (ByteBufferPipe.this.tail = 0);
                    }
                };

                this.outputStream = new OutputStream() {

                    @Override public void
                    write(int b) throws IOException {
                        this.write(new byte[] { (byte) b }, 0, 1);
                    }

                    @NotNullByDefault(false) @Override public void
                    write(byte[] ba, int off, int len) throws IOException {
                        ByteBufferPipe.this.write(ba, off, len);
                    }

                    @Override public void
                    close() {
                        ByteBufferPipe.this.closed = true;
                    }
                };
            }

            /**
             * ???
             */
            public int
            read(byte[] ba, int off, int len) throws IOException {
                if (off < 0 || off > ba.length) throw new IllegalArgumentException("off");
                if (len < 0 || off + len > ba.length) throw new IllegalArgumentException("len");

                if (len == 0) return 0;

                int n;
                if (blockingRead) {
                    n = makeAllocated();
                } else {
                    n = getAllocated();
                    if (n == 0) throw new BlockingException();
                }
                if (n == -1) return -1;
                if (n > len) n = len;

                byteBuffer.position(ByteBufferPipe.this.head);
                if (ByteBufferPipe.this.head + n < ByteBufferPipe.this.capacity) {
                    byteBuffer.get(ba, off, len);
                    ByteBufferPipe.this.head += n;
                } else {
                    int x = ByteBufferPipe.this.capacity - ByteBufferPipe.this.head;
                    byteBuffer.get(ba, off, x);
                    byteBuffer.position(0);
                    byteBuffer.get(ba, off + x, n - x);
                    ByteBufferPipe.this.head = n - x;
                }
                if (ByteBufferPipe.this.head == ByteBufferPipe.this.tail) ByteBufferPipe.this.empty = true;
                synchronized (ByteBufferPipe.this.notFull) {
                    ByteBufferPipe.this.notFull.notifyAll(); // Notify any producers that there is space.
                }
                return n;
            }

            /**
             * ???
             */
            protected void
            write(byte[] ba, int off, int len) throws IOException {
                if (off < 0 || off > ba.length) throw new IllegalArgumentException("off");
                if (len < 0 || off + len > ba.length) throw new IllegalArgumentException("len");

                if (ByteBufferPipe.this.closed) throw new IOException("closed");

                while (len > 0) {

                    int n;
                    if (blockingWrite) {
                        n = makeFree();
                    } else {
                        n = getFree();
                        if (n < len) throw new BlockingException();
                    }
                    if (n > len) n = len;

                    byteBuffer.position(ByteBufferPipe.this.tail);
                    if (ByteBufferPipe.this.tail + n < ByteBufferPipe.this.capacity) {
                        byteBuffer.put(ba, off, n);
                        ByteBufferPipe.this.tail += n;
                    } else {
                        int x = ByteBufferPipe.this.capacity - ByteBufferPipe.this.tail;
                        byteBuffer.put(ba, off, x);
                        byteBuffer.position(0);
                        byteBuffer.put(ba, off + x, n - x);
                        ByteBufferPipe.this.tail = n - x;
                    }
                    len                       -= n;
                    off                       += n;
                    ByteBufferPipe.this.empty = false;
                    synchronized (ByteBufferPipe.this.notEmpty) {
                        ByteBufferPipe.this.notEmpty.notifyAll(); // Notify any consumers that there is data.
                    }
                }
            }

            /**
             * @return The number of ALLOCATED bytes in the ring buffer (1 or more), or -1 to indicate end-of-input
             */
            private int
            makeAllocated() throws InterruptedIOException {
                synchronized (this.notEmpty) {
                    for (;;) {
                        int allocated = getAllocated();
                        if (allocated != 0) return allocated;
                        try {
                            this.notEmpty.wait();
                        } catch (InterruptedException e) {
                            throw new InterruptedIOException(); // SUPPRESS CHECKSTYLE Avoid Hiding Cause of the Exception
                        }
                    }
                }
            }

            /**
             * @return The number of ALLOCATED bytes in the ring buffer (0 or more), or -1 to indicate end-of-input
             */
            protected int
            getAllocated() {
                return (
                    this.empty ? (this.closed ? -1 : 0) :
                    this.head < this.tail ? this.tail - this.head :
                    this.capacity + this.tail - this.head
                );
            }

            /**
             * @return The number of FREE bytes in the ring buffer; >= 1
             */
            private int
            makeFree() throws InterruptedIOException {
                synchronized (ByteBufferPipe.this.notFull) {
                    for (;;) {
                        int av = getFree();
                        if (av > 0) return av;

                        if (this.closed) return -1;

                        try {
                            ByteBufferPipe.this.notFull.wait();
                        } catch (InterruptedException e) {
                            throw new InterruptedIOException(); // SUPPRESS CHECKSTYLE Avoid Hiding Cause of the Exception
                        }
                    }
                }
            }

            /**
             * @return The number of FREE bytes in the ring buffer; >= 0
             */
            private int
            getFree() {
                return (
                    ByteBufferPipe.this.empty
                    ? ByteBufferPipe.this.capacity
                    : ByteBufferPipe.this.head >= ByteBufferPipe.this.tail
                    ? ByteBufferPipe.this.head - ByteBufferPipe.this.tail
                    : ByteBufferPipe.this.capacity - ByteBufferPipe.this.tail + ByteBufferPipe.this.head
                );
            }
        }

        return new ByteBufferPipe();
    }

    /**
     * Creates and returns a {@link Pipe} which enlarges its {@link ByteBuffer} as needed by invoking the
     * {@code byteBufferProducer}. Thus wrte operations on the pipe will never block.
     */
    public static Pipe
    elasticPipe(
        final Transformer<Integer, ByteBuffer> byteBufferProducer,
        final int                              initialCapacity,
        final boolean                          blockingRead
    ) {

        /**
         * @see #elasticPipe
         */
        class ElastcPipe implements Pipe {

            private ByteBuffer              byteBuffer        = byteBufferProducer.transform(initialCapacity);
            private Pipe                    pipe              = byteBufferPipe(this.byteBuffer, blockingRead, false);
            private final ProxyInputStream  proxyInputStream  = new ProxyInputStream(this.pipe.getInputStream()) {};
            private final ProxyOutputStream proxyOutputStream = new ProxyOutputStream(this.pipe.getOutputStream()) {

                @Override public void
                write(int b) throws IOException {
                    this.write(new byte[] { (byte) b }, 0, 1);
                }

                @NotNullByDefault(false) @Override public void
                write(byte[] ba, int off, int len) throws IOException {
                    try {
                        super.write(ba, off, len);
                        return;
                    } catch (BlockingException be) {
                        int capacity = ElastcPipe.this.byteBuffer.capacity();
                        int x;
                        for (x = 2 * capacity; x < capacity - len; x *= 2);
                        ByteBuffer newByteBuffer = byteBufferProducer.transform(x);
                        Pipe       newPipe       = byteBufferPipe(newByteBuffer, blockingRead, false);
                        try {
                            IoUtil.copy(ElastcPipe.this.pipe.getInputStream(), newPipe.getOutputStream());
                        } catch (BlockingException be2) {
                            ;
                        }
                        newPipe.getOutputStream().write(ba, off, len);
                        ElastcPipe.this.byteBuffer = newByteBuffer;
                        ElastcPipe.this.pipe       = newPipe;
                        ElastcPipe.this.proxyInputStream.setDelegate(newPipe.getInputStream());
                        ElastcPipe.this.proxyOutputStream.setDelegate(newPipe.getOutputStream());
                    }
                }
            };

            @Override public InputStream
            getInputStream()  { return this.proxyInputStream; }

            @Override public OutputStream
            getOutputStream() { return this.proxyOutputStream; }
        }
        return new ElastcPipe();
    }
}
