
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.lang.protocol;

/**
 * Various {@link Predicate}-related utility methods.
 */
public final
class PredicateUtil {

    private
    PredicateUtil() {}

    /**
     * @return A {@link Predicate} who's {@link Predicate#evaluate(Object)} method always returns {@code true}.
     */
    @SuppressWarnings("unchecked") public static <T> Predicate<T>
    always() {
        return ALWAYS;
    }

    /**
     * @return A {@link Predicate} who's {@link Predicate#evaluate(Object)} method always returns {@code false}.
     */
    @SuppressWarnings("unchecked") public static <T> Predicate<T>
    never() {
        return NEVER;
    }

    @SuppressWarnings("rawtypes") private static final Predicate
    ALWAYS = new Predicate() {

        @Override public boolean
        evaluate(Object subject) { return true; }

        @Override public String
        toString() { return "ALWAYS"; }
    };

    @SuppressWarnings("rawtypes") private static final Predicate
    NEVER = new Predicate() {

        @Override public boolean
        evaluate(Object subject) { return false; }

        @Override public String
        toString() { return "NEVER"; }
    };

    /**
     * @return A {@link Predicate} which returns {@code true} iff both {@code p1} and {@code p2} return {@code true}
     *         for any given {@code subject}.
     */
    @SuppressWarnings("unchecked") public static <T> Predicate<T>
    and(final Predicate<? super T> p1, final Predicate<? super T> p2) {
        if (p1 == NEVER || p2 == NEVER) return NEVER;
        if (p1 == ALWAYS) return (Predicate<T>) p2;
        if (p2 == ALWAYS) return (Predicate<T>) p1;

        return new Predicate<T>() {

            @Override public boolean
            evaluate(T subject) { return p1.evaluate(subject) && p2.evaluate(subject); }

            @Override public String
            toString() { return p1 + " && " + p2; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} if {@code p1} or {@code p2} returns {@code true}
     *         for a given {@code subject}.
     */
    @SuppressWarnings("unchecked") public static <T> Predicate<T>
    or(final Predicate<? super T> p1, final Predicate<? super T> p2) {
        if (p1 == ALWAYS || p2 == ALWAYS) return ALWAYS;
        if (p1 == NEVER) return (Predicate<T>) p2;
        if (p2 == NEVER) return (Predicate<T>) p1;

        return new Predicate<T>() {

            @Override public boolean
            evaluate(T subject) { return p1.evaluate(subject) || p2.evaluate(subject); }

            @Override public String
            toString() { return p1 + " || " + p2; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff the {@code delegate} returns {@code false} for any
     *         given {@code subject}.
     */
    public static <T> Predicate<T>
    not(final Predicate<T> delegate) {
        return new Predicate<T>() {

            @Override public boolean
            evaluate(T subject) { return !delegate.evaluate(subject); }

            @Override public String
            toString() { return "!" + delegate; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is less than the
     *         {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    less(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) < 0; }

            @Override public String
            toString() { return "< " + other; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is less than or equal to the
     *         {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    lessEqual(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) <= 0; }

            @Override public String
            toString() { return "<= " + other; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is greater than the
     *         {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    greater(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) > 0; }

            @Override public String
            toString() { return "> " + other; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is greater than or equal to
     *         the {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    greaterEqual(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) >= 0; }

            @Override public String
            toString() { return ">= " + other; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is (arithmetically) equal
     *         to the {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    equal(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) == 0; }

            @Override public String
            toString() { return "== " + other; }
        };
    }

    /**
     * @return A {@link Predicate} which returns {@code true} iff a given {@code subject} is (arithmetically) not equal
     *         to the {@code other}.
     */
    public static <C extends Comparable<C>> Predicate<C>
    notEqual(final C other) {
        return new Predicate<C>() {

            @Override public boolean
            evaluate(C subject) { return subject.compareTo(other) != 0; }

            @Override public String
            toString() { return "!= " + other; }
        };
    }

    /**
     * @return {@code min <= subject <= max}, or, if {@code min > max}: {@code subject >= min || subject <= max}
     */
    public static <C extends Comparable<C>> Predicate<C>
    between(final C min, final C max) {
        return (
            max.compareTo(min) >= 0
            ? new Predicate<C>() {

                @Override public boolean
                evaluate(C subject) { return subject.compareTo(min) >= 0 && subject.compareTo(max) <= 0; }

                @Override public String
                toString() { return min + "-" + max; }
            }
            : new Predicate<C>() {

                @Override public boolean
                evaluate(C subject) { return subject.compareTo(min) <= 0 || subject.compareTo(max) >= 0; }

                @Override public String
                toString() { return min + "-" + max; }
            }
        );
    }

    /** Value equality */
    public static <T> Predicate<T>
    equals(final T other) {
        return new Predicate<T>() {

            @Override public boolean
            evaluate(T subject) { return subject.equals(other); }

            @Override public String
            toString() { return "eq " + other; }
        };
    }

    /** Value equality */
    public static <T> Predicate<T>
    notEquals(final T other) {
        return new Predicate<T>() {

            @Override public boolean
            evaluate(T subject) { return !subject.equals(other); }

            @Override public String
            toString() { return "nq " + other; }
        };
    }

    /**
     * Converts a {@link Predicate} (the "source") into a different {@link Predicate} (the "target"), which is possible
     * iff:
     * <ul>
     *   <li>The source's element type is a subclass of the target's element type, and
     *   <li>The source's exception is a subclass of the target's exception.
     * </ul>
     *
     * @param <ST> The source predicate's element type
     * @param <TT> The target predicate's element type
     */
    @SuppressWarnings({ "unchecked", "rawtypes" }) public static <ST, TT extends ST> Predicate<TT>
    widen(final Predicate<ST> source) {
        return (Predicate) source;
    }

    /**
     * Converts a {@link PredicateWhichThrows} (the "source") into a different {@link PredicateWhichThrows} (the
     * "target"), which is possible iff:
     * <ul>
     *   <li>The source's element type is a subclass of the target's element type, and
     *   <li>The source's exception is a subclass of the target's exception.
     * </ul>
     *
     * @param <ST> The source predicate's element type
     * @param <TT> The target predicate's element type
     * @param <TE> The target predicate's exception
     * @param <SE> The source predicate's exception
     */
    // SUPPRESS CHECKSTYLE Maximum Line Length
    @SuppressWarnings({ "rawtypes", "unchecked" }) public static <ST, TT extends ST, TE extends Throwable, SE extends TE> PredicateWhichThrows<TT, TE>
    widen(final PredicateWhichThrows<ST, SE> source) {
        return (PredicateWhichThrows) source;
    }

    /**
     * Converts a {@link Predicate} into a {@link PredicateWhichThrows}, which is possible iff the source's element
     * type is a subclass of the target's element type.
     *
     * @param <ST> The source predicate's element type
     * @param <TT> The target predicate's element type
     * @param <TE> The target predicate's exception
     */
    public static <ST, TT extends ST, TE extends Throwable> PredicateWhichThrows<TT, TE>
    cast(final Predicate<ST> source) {
        return new PredicateWhichThrows<TT, TE>() {

            @Override public boolean
            evaluate(TT subject) { return source.evaluate(subject); }
        };
    }
}
