
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.text.expression;

import de.unkrig.commons.nullanalysis.Nullable;

/**
 * An expression is something that evaluates to a value each time it is {@link #evaluate(Object[]) evaluated}.
 * If you want to evaluate an expression only once, you better use {@link
 * de.unkrig.commons.text.expression.ExpressionEvaluator#evaluate(String, Object[])}.
 * <p>
 * To parse an {@link Expression} from a string, use {@link de.unkrig.commons.text.expression.Parser#parse()}:
 * <pre>
 *     import de.unkrig.commons.text.expression.*;
 *
 *     Expression e = new ExpressionEvaluator(new String[] { "a", "b" }).parse("a + b");
 *     Object o = e.evaluate(new Object[] { 3, 4 });
 * </pre>
 */
public
interface Expression {

    /**
     * Computes the value of the expression
     *
     * @param variableValues The values of the variables that were named when the expression was parsed. Allowed values
     *                       are {@link String}, {@link Integer}, {@link Boolean}, any other {@code Object}, or {@code
     *                       null}.
     * @return               A {@link String}, {@link Integer}, {@link Boolean}, any other {@code Object}, or {@code
     *                       null}
     */
    @Nullable Object
    evaluate(Object[] variableValues) throws EvaluationException;

    /**
     * Shorthand for {@code new String[0]}.
     */
    String[] NO_VARIABLE_NAMES = new String[0];
    
    /**
     * Shorthand for {@code new Object[0]}.
     */
    Object[] NO_VARIABLE_VALUES = new Object[0];
}
