
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.text.pattern;

import static de.unkrig.commons.text.pattern.Pattern2.WILDCARD;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.unkrig.commons.nullanalysis.Nullable;

/**
 * @see #matches
 */
public final
class IncludeExclude extends Glob {

    /**
     * File names that one would typically exclude from file searches.
     */
    public static final Glob DEFAULT_EXCLUDES;
    static {
        IncludeExclude ie = new IncludeExclude();
        ie.appendExclude(Glob.compile(".svn",     WILDCARD));
        ie.appendExclude(Glob.compile("***/.svn", WILDCARD));
        ie.appendExclude(Glob.compile("***!.svn", WILDCARD));
        ie.appendExclude(Glob.compile("CVS",      WILDCARD));
        ie.appendExclude(Glob.compile("***/CVS",  WILDCARD));
        ie.appendExclude(Glob.compile("***!CVS",  WILDCARD));
        DEFAULT_EXCLUDES = ie;
    }

    public IncludeExclude() {}

    /**
     * Takes effect iff none of the previously added includes and excludes match.
     * @see #matches(String)
     * @see #replace(String)
     */
    public void
    appendInclude(Glob include) {
        this.mags.add(new ModeAndGlob(Mode.INCLUDE, include));
    }

    /**
     * Takes precendence over all previously added includes and excludes.
     * @see #matches(String)
     * @see #replace(String)
     */
    public void
    prependInclude(Glob include) {
        this.mags.add(0, new ModeAndGlob(Mode.INCLUDE, include));
    }

    /**
     * Takes effect iff none of the previously added includes and excludes match.
     * @see #matches(String)
     * @see #replace(String)
     */
    public void
    appendExclude(Glob exclude) {
        this.mags.add(new ModeAndGlob(Mode.EXCLUDE, exclude));
    }

    /**
     * Takes precendence over all previously added includes and excludes.
     * @see #matches(String)
     * @see #replace(String)
     */
    public void
    prependExclude(Glob exclude) {
        this.mags.add(0, new ModeAndGlob(Mode.EXCLUDE, exclude));
    }

    /**
     * All includes and excludes are matched against the {@code subject}. On the first successful match,
     * {@code true} or {@code false} is returned, depending on whether the match was an INCLUDE or an EXCLUDE.
     * If none of the includes and excludes match, {@code true} or {@code false} is returned, depending on whether the
     * lowest-priority rule is an EXCLUDE or an INCLUDE.
     */
    @Override public boolean
    matches(String subject) {

        if (this.mags.isEmpty()) return true;

        boolean isContainer = subject.endsWith("!") || subject.endsWith("/");

        for (ModeAndGlob mag : this.mags) {
            if (mag.mode == Mode.INCLUDE) {
                if (mag.glob.matches(subject)) return true;
            } else {
                if (!isContainer && mag.glob.matches(subject)) return false;
            }
        }
        return this.mags.get(this.mags.size() - 1).mode == Mode.EXCLUDE;
    }

    /**
     * The {@code subject} is transformed by all matching includes, until an exclude matches.
     *
     * @return The transformed subject, or {@code null} iff an exclude matches before the first include matches
     */
    @Override @Nullable public String
    replace(String subject) {
        boolean hadInclude = false;
        for (ModeAndGlob map : this.mags) {
            switch (map.mode) {
            case INCLUDE:
                String tmp = map.glob.replace(subject);
                if (tmp != null) {
                    subject    = tmp;
                    hadInclude = true;
                }
                break;
            case EXCLUDE:
                if (map.glob.matches(subject)) return hadInclude ? subject : null;
                break;
            default:
                throw new IllegalStateException();
            }
        }
        return hadInclude ? subject : null;
    }

    @Override public String
    toString() {
        Iterator<ModeAndGlob> it = this.mags.iterator();
        if (!it.hasNext()) return "(empty)";

        StringBuilder sb = new StringBuilder();

        {
            ModeAndGlob mag = it.next();
            if (mag.mode == Mode.EXCLUDE) sb.append('~');
            sb.append(mag.glob);
        }

        while (it.hasNext()) {
            ModeAndGlob mag = it.next();
            sb.append(mag.mode == Mode.INCLUDE ? ',' : '~');
            sb.append(mag.glob);
        }

        return sb.toString();
    }

    private static enum Mode { INCLUDE, EXCLUDE }

    private static
    class ModeAndGlob {
        final Mode mode;
        final Glob glob;

        ModeAndGlob(Mode mode, Glob glob) {
            this.mode = mode;
            this.glob = glob;
        }
    }

    private final List<ModeAndGlob> mags = new ArrayList<ModeAndGlob>();
}
