
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.util.tool.logging;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Filter;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;

import de.unkrig.commons.nullanalysis.Nullable;

// CHECKSTYLE TypeName:OFF       Type names
// CHECKSTYLE VariableCheck:OFF  Field JAVADOC
// CHECKSTYLE MethodCheck:OFF    Method JAVADOC

/**
 * An <a href="http://ant.apache.org">ANT</a> task which configures the JDK logging {@code java.util.logging}.
 */
public
class AntTask extends Task {

    private final List<Names_Filter_Level_Parent_UseParentHandlers_Handlers_Element> loggerElements = (
        new ArrayList<AntTask.Names_Filter_Level_Parent_UseParentHandlers_Handlers_Element>()
    );

    // BEGIN CONFIGURATION SETTERS

    /**
     * An element with a 'value' attribute and text.
     */
    public static
    class Value_Element {

        @Nullable String value;

        public void
        setValue(String value) { this.value = value; }

        public void
        addText(@Nullable String text) {
            if (text == null) return;
            if (this.value != null) {
                throw new RuntimeException("'value' attribute and element text are mutually exclusive");
            }
            this.value = text;
        }
    }

    /**
     * An element with two attributes 'name' and 'value', and text.
     */
    public static
    class Name_Value_Element extends Value_Element {

        @Nullable String name;

        public void
        setName(String name) { this.name = name; }
    }

    /**
     * An element with
     * attributes 'className' and 'argument',
     * subelements 'argument' and 'attribute',
     * and text.
     */
    public static
    class ClassName_Arguments_Attributes_Element {
        @Nullable String          className;
        final List<String>        arguments  = new ArrayList<String>();
        final Map<String, String> attributes = new HashMap<String, String>();

        public void
        setClassName(String className) { this.className = className; }

        public void
        setArgument(String value) { this.arguments.add(value); }

        public void
        addConfiguredArgument(Value_Element element) {
            if (element.value == null) {
                throw new IllegalArgumentException("'value' attribute of element 'argument' is missing");
            }
            this.arguments.add(element.value);
        }

        public void
        addConfiguredAttribute(Name_Value_Element element) {
            if (element.name == null) {
                throw new IllegalArgumentException("'name' attribute of element 'attribute' is missing");
            }
            if (element.value == null) {
                throw new IllegalArgumentException("'value' attribute of element 'attribute' is missing");
            }
            this.attributes.put(element.name, element.value);
        }

        Object
        instantiate() throws Exception {
            if (this.className == null) throw new IllegalArgumentException("'classname' attribute is missing");
            Class<?> clasS = Class.forName(this.className);

            // Find the applicable constructor.
            Constructor<?> constructor;
            CONSTRUCTORS:
            {
                for (Constructor<?> c : clasS.getConstructors()) {
                    Class<?>[] parameterTypes = c.getParameterTypes();
                    if (parameterTypes.length == this.arguments.size()) {
                        constructor = c;
                        break CONSTRUCTORS;
                    }
                }
                throw new NoSuchMethodException(this.className + this.arguments);
            }

            // Instantiate the object.
            Object object;
            {
                Class<?>[] parameterTypes = constructor.getParameterTypes();
                Object[]   tmp            = new Object[parameterTypes.length];
                for (int i = 0; i < tmp.length; i++) {
                    Class<?> parameterType = parameterTypes[i];
                    Object   argument      = this.arguments.get(i);
                    if (parameterType != String.class) {
                        argument = parameterType.getConstructor(String.class).newInstance(argument);
                    }
                    tmp[i] = argument;
                }
                object = constructor.newInstance(tmp);
            }

            // Invoke setters for the given attributes.
            for (Entry<String, String> entry : this.attributes.entrySet()) {
                String name  = entry.getKey();
                Object value = entry.getValue();

                String methodName = "set" + Character.toUpperCase(name.charAt(0)) + name.substring(1);

                // Find the applicable setter method.
                Method   method;
                Class<?> parameterType;
                METHODS:
                {
                    for (Method m : clasS.getDeclaredMethods()) {
                        if (!m.getName().equals(methodName)) continue;
                        Class<?>[] parameterTypes = m.getParameterTypes();
                        if (parameterTypes.length == 1) {
                            method        = m;
                            parameterType = parameterTypes[0];
                            break METHODS;
                        }
                    }
                    throw new NoSuchMethodException(this.className + "::" + methodName);
                }

                // Invoke the setter.
                if (parameterType != String.class) {
                    value = parameterType.getConstructor(String.class).newInstance(value);
                }
                method.invoke(object, value);
            }
            return object;
        }
    }

    /**
     * An element with
     * attributes 'className' and 'argument',
     * subelements 'argument', 'attribute', 'filter' and 'formatter',
     * and text.
     */
    public static
    class ClassName_Arguments_Attributes_Filter_Formatter_Element extends ClassName_Arguments_Attributes_Element {
        @Nullable Filter    filter;
        @Nullable Formatter formatter;

        public void
        addConfiguredFilter(ClassName_Arguments_Attributes_Element element) throws Exception {
            if (this.filter != null) throw new RuntimeException("At most one 'filter' subelement allowed");
            this.filter = (Filter) element.instantiate();
        }

        public void
        addConfiguredFormatter(ClassName_Arguments_Attributes_Element element) throws Exception {
            if (this.formatter != null) throw new RuntimeException("At most one 'formatter' subelement allowed");
            this.formatter = (Formatter) element.instantiate();
        }
    }

    /**
     * An element with
     * attributes 'name', 'clearLevel', 'level', 'parent', 'useParentHandlers' and 'clearHandlers',
     * subelements 'name', ''filter' and 'handler',
     * and text.
     */
    public static
    class Names_Filter_Level_Parent_UseParentHandlers_Handlers_Element {
        final List<String>  names = new ArrayList<String>();
        public boolean      clearLevel;
        @Nullable Level     level;
        @Nullable Logger    parent;
        @Nullable Boolean   useParentHandlers;
        public boolean      clearFilter;
        @Nullable Filter    filter;
        boolean             clearHandlers;
        final List<Handler> handlers = new ArrayList<Handler>();

        public void
        setName(String name) { this.names.add(name); }

        public void
        addConfiguredName(Value_Element element) {
            if (element.value == null) {
                throw new IllegalArgumentException("'value' attribute of element 'name' is missing");
            }
            this.names.add(element.value);
        }

        public void
        setClearLevel(boolean value) { this.clearLevel = value; }

        public void
        setLevel(String value) { this.level = Level.parse(value); }

        public void
        setParent(String value) { this.parent = Logger.getLogger(value); }

        public void
        setUseParentHandlers(Boolean value) { this.useParentHandlers = value; }

        public void
        addConfiguredFilter(ClassName_Arguments_Attributes_Element element) throws Exception {
            if (this.filter != null) throw new RuntimeException("At most one 'filter' subelement allowed");
            this.filter = (Filter) element.instantiate();
        }

        public void
        setClearHandlers(boolean value) { this.clearHandlers = value; }

        public void
        addConfiguredHandler(ClassName_Arguments_Attributes_Filter_Formatter_Element element) throws Exception {
            Handler handler = (Handler) element.instantiate();
            handler.setFilter(element.filter);
            if (element.formatter != null) handler.setFormatter(element.formatter);
            this.handlers.add(handler);
        }
    }

    public void
    addConfiguredLogger(Names_Filter_Level_Parent_UseParentHandlers_Handlers_Element element) {
        this.loggerElements.add(element);
    }

    // END CONFIGURATION SETTERS

    /**
     * The ANT task "execute" method.
     *
     * @see Task#execute()
     */
    @Override public void
    execute() throws BuildException {
        try {

            for (Names_Filter_Level_Parent_UseParentHandlers_Handlers_Element element : this.loggerElements) {
                List<String> loggerNames = element.names;
                if (loggerNames.isEmpty()) loggerNames = Collections.singletonList("");
                for (String name : loggerNames) {
                    Logger logger = Logger.getLogger(name);
                    if (element.clearFilter) logger.setFilter(null);
                    if (element.filter != null) logger.setFilter(element.filter);
                    if (element.clearLevel) logger.setLevel(null);
                    if (element.level != null) logger.setLevel(element.level);
                    if (element.parent != null) logger.setParent(element.parent);
                    if (element.useParentHandlers != null) logger.setUseParentHandlers(element.useParentHandlers);
                    // It DOES make sense to specify zero handlers: If only filter, level, parent and/or
                    // useParentHandlers are to be reconfigured.
                    if (element.clearHandlers) {
                        for (Handler handler : logger.getHandlers()) logger.removeHandler(handler);
                    }
                    for (Handler handler : element.handlers) logger.addHandler(handler);
                }
            }
        } catch (Exception e) {
            throw new BuildException(e);
        }
    }
}
